#!/usr/local/bin/zasm --reqcolon --dotnames -o original/

#target ram

stack_size:	equ	0x400
data_size:  equ 0x100
STDIN  = 0xFF
STDOUT = 0xFF

enums = 0
NOERROR = enums++
REFCNT_OVERFLOW = enums++
OUT_OF_MEMORY = enums++


#code PAGE0
#code ROP		; ATTN: 256/32-byte pages which intersect ROP should/must be immutable
#code CODE		; immutable code & const data
#code RAM		; mutable code & initialized variables
#code GSINIT	; initialization code. may be added to HEAP
#code HEAP
	heap_start:	equ (HEAP+1)&$FFFE;GSINIT
	heap_end: 	equ (DATA+0)&$FFFE
#data DATA, STACK - data_size, data_size 	; zero-initialized data
#data STACK, 0x10000-stack_size		; VM / user stack


.macro putc			; mcode, pres: all
	out (STDOUT),a
.endm

.macro getc			; mcode, out: a=char
	in a,(STDIN)
.endm

.macro msg &CSTRING	; mcode, mod: af hl
	rst putmsg
	.db	&CSTRING, 0
.endm

.macro call_ix		; mcode
	rst jp_ix
.endm

.macro panic &errno	; mcode
	ld  a,&errno
	rst	jp_panic
.endm

; --------------------------------
; clear the zero-initialized DATA area

.area GSINIT
	.assert DATA_size > 1
	ld	bc,DATA_size - 1
	ld	de,DATA + 1
	ld	hl,DATA
	ld	(hl),0
	ldir

; ---------------------------
; Interrupt data

.area DATA
sys_tick:	defs 4
sys_stack:	defs 10		; adjust as needed
sys_stack_end:
sys_usp:	defs 2


; ---------------------------
; get sys_tick:

.area CODE
systick::		; ( -- n16 )
	ld	hl,(sys_tick)
	ret
systick.l::		; ( -- n32 )
	ld	hl,(sys_tick)
	exx
	ld	hl,(sys_tick+2)
	exx
	ld	a,(sys_tick)
	cp	l
	ret	z
	jr	systick.l



; ---------------------------
; include everything
; ---------------------------

#include "vm-macros.s"
#include "vm-opcodes.s"
#include "vm-heap.s"
#include "vm-lib.s"
#include "vm-core.s"
#include "vm-int.s"
#include "vm-int32.s"
#include "vm-rop-repair.s"
#include "test.s"


; ---------------------------
; Restart Vectors
; ---------------------------

.area PAGE0
.org 0
reset:			; ( -- )
  di
  ld  sp,0
  jp  GSINIT

.org 8
putc:			; ( a:char -- )			for native mode!
  out (STDOUT),a
  ret

.org 16
getc:			; ( -- a:char f:z )		for native mode!
  in  a,(STDIN)
  and a
  ret

.org 24 - 2
  	out (STDOUT),a
putmsg:			; ( pc:cstring -- )		for native mode!	mod: hl, af
  	pop hl
	ld  a,(hl++)
	push hl
	and	a
	jr	nz,putmsg-2
	ret

.org 32
jp_ix:
	jp  ix

.org 40
jp_panic:		; ( a:errno -- )
  jp  panic

.org 48
  jp  $


; ---------------------------------------
; system's interrupt handler   [39 bytes]
; entered with return addr possibly written into ROP code
; -> switch to system stack
;	 do whatever we need to do
;	 test whether the return address was written into ROP code
;	 -> repair word & exit
;	 else switch back to user stack
;	 ei & ret

.org 56
interrupt_handler:

; switch to "system stack" and save registers
; note: we must not overwrite more ROP words!

	ld 	(sys_usp),sp		; save user stack pointer
	ld	sp,sys_stack_end	; use system stack
	push af
	push bc
	push de
	push hl

; main interrupt action:
; increment system time
	ld	hl,sys_tick
1$:	inc	(hl++)
	jr	z,1$

; test whether we overwrote ROP code:
	ld	hl,(sys_usp)
	ld	a,h
	sub	a,hi(rop_codestart)
	sub	a,hi(rop_codeend+0xff)-hi(rop_codestart)
jp_rop_repair:: 		; patched by some tests
	jp	c, rop_repair

; pop registers, switch back to "user stack" and return from interrupt
	pop	hl
	pop	de
	pop	bc
	pop	af
	ld	sp,(sys_usp)	; if code is in ram then sys_usp could be here at $+1
	ei
	ret


; --------------------------------
; Non-maskable Interrupt
;
; if an NMI interrupts a normal interrupt then we are lost.
; the current ROP code repair mechanism can't repair 2 words.
; So there is no support for NMI.

#if 0
.org 0x66
	rst 0
#endif



.area CODE
panic_vm::		; ( [errno<<8] -- )
	pop	af
panic::			; ( a:errno -- )
	di
	ld	(-2),hl	; hl
	pop	hl
	ld	(-4),sp	; sp
	ld	sp,-4
	push hl		; pc


	push af
	msg	<10,"PANIC: errno=$">
	pop	af
	ld	l,a
	call puthex.b

	msg  <10,"ip=$">
	pop  hl
	call puthex

	msg  <10,"sp=$">
	pop	 hl
	call puthex

	msg  <10,"iy=$">
	push iy
	pop	 hl
	call puthex

	msg  <10,"hl=$">
	pop	 hl
	call puthex

	msg  <10>
	jr	 $

puthex.l::		; ( u32 -- )
	exx
	call puthex
	exx
puthex::		; ( u16 -- )
	push hl
	ld	l,h
	call puthex.b
	pop hl
puthex.b::		; ( u8 -- )			mod: af, pres all other		uses stack
	ld	a,l
	rra \ rra \ rra \ rra
	call put_hex_digit
	ld	a,l
put_hex_digit:	; ( a:u8 -- )		mod: af, pres all other
	and	0x0f
	add	'0'
	cp	'9'+1
	jr	c,1$
	add	7
1$:	jp	putc


; --------------------------------
; align end of page0 to a multiple of 2
; the ROP segment follows and ROP opcodes must be word-aligned
;
; actually the page0 code currently happens to be $5F bytes long
; thus the first ROP code page would start at $40
; which includes most of the interrupt handler.
; the test code pokes into the interrupt handler
; and then the checksum tests in the test code fail.
;
; note: the padding to multiples of $100 at both ends of the ROP segment
; must be either immutable OR the sp must never point there,
; except for the padding to full rop pages, currently 32 bytes.
; as the sp (hopefully) never points into page0 we can modify data in here.

.area PAGE0
.align 2



; ---------------------------
; at end of GSINIT jump to main program:

.area GSINIT
	ld	sp,0	; just in case
	im 1		; always RST 56
	ei
	call main
main:: 			; there is no application
	jp $		; this jump is patched by some tests!



; ---------------------------
; resolve local variables etc.

rop_final_resolver



